#include "windows.h"
#include "Python.h"

static int string_width(char *str)
{
    return MultiByteToWideChar(CP_ACP,         /* code page */
			       MB_PRECOMPOSED, /* character-type options */
			       str, -1, (LPWSTR)"\0", 0);
}

static int string_convert(char *str, char *ptr, int len)
{
    return MultiByteToWideChar(CP_ACP,         /* code page */
			       MB_PRECOMPOSED, /* character-type options */
			       str, -1,
			       (LPWSTR)ptr, len);
}

static int dlg_header(char *data,
		      int style, int x, int y, int cx, int cy,
		      int num_items,
		      char *menuname,
		      char *classname,
		      char *caption,
		      int points, char *fontname)
{
    int block_len, menuname_len, classname_len, caption_len, fontname_len;
    char *ptr;
    DLGTEMPLATE *dlg;

    if (menuname == NULL)
	menuname = "";
    if (classname == NULL)
	classname = "";

    menuname_len = string_width(menuname);
    classname_len = string_width(classname);
    caption_len = string_width(caption);

    block_len = sizeof(DLGTEMPLATE)
	+ (menuname_len + classname_len + caption_len) * 2;
    if (fontname != NULL)
	style |= DS_SETFONT;

    /* if DS_SETFONT then need to include font desc */
    if (style & DS_SETFONT) {
	block_len += sizeof(WORD);
	fontname_len = string_width(fontname);
	block_len += fontname_len * 2;
    } else
	fontname_len = 0;

    if (data == NULL)
	return block_len;

    memset(data, 0, block_len);

    dlg = (DLGTEMPLATE*)data;
    dlg->style = style;
    dlg->cdit = num_items;
    dlg->x = x;
    dlg->y = y;
    dlg->cx = cx;
    dlg->cy = cy;

    ptr = (char*)(dlg + 1);
    string_convert(menuname, ptr, menuname_len);
    ptr += menuname_len * 2;
    string_convert(classname, ptr, classname_len);
    ptr += classname_len * 2;
    string_convert(caption, ptr, caption_len);
    ptr += caption_len * 2;

    if (style & DS_SETFONT) {
	*(WORD*)ptr = points;
	ptr += sizeof(WORD);
	string_convert(fontname, ptr, fontname_len);
    }

    return block_len;
}

static int dlg_item(char *data,
		    int style, int x, int y, int cx, int cy, int id,
		    char *classname, int predefined,
		    char *caption)
{
    int block_len, classname_len, caption_len;
    char *ptr;
    DLGITEMTEMPLATE *item;

    if (classname != NULL)
	classname_len = string_width(classname);
    else
	classname_len = 2;
    if (caption == NULL)
	caption = "";
    caption_len = string_width(caption);

    block_len = sizeof(DLGITEMTEMPLATE)
	+ (classname_len + caption_len + 1) * 2;

    if (data == NULL)
	return block_len;

    memset(data, 0, block_len);

    item = (DLGITEMTEMPLATE*)data;
    item->style = style;
    item->x = x;
    item->y = y;
    item->cx = cx;
    item->cy = cy;
    item->id = id;

    ptr = (char*)(item + 1);

    if (classname != NULL) {
	string_convert(classname, ptr, classname_len);
	ptr += classname_len * 2;
    } else {
	*(WORD*)ptr = 0xffff;
	ptr += sizeof(WORD);
	*(WORD*)ptr = predefined;
	ptr += sizeof(WORD);
    }
    string_convert(caption, ptr, caption_len);
    ptr += caption_len * 2;

    *(WORD*)ptr = 0;

    return block_len;
}

static int align(int len, int size)
{
    return ((len + size - 1) / size) * size;
}

static int parse_dlg_item(char *data, int block_len, PyObject *seq, int i)
{
    PyObject *obj, *tup;
    PyObject *classname_obj, *caption_obj;
    char *classname, *caption;
    int res;
    int id, x, y, cx, cy, style, predefined;

    block_len = align(block_len, 4);
    if (data != NULL)
	data += block_len;

    obj = PySequence_GetItem(seq, i + 1);
    if (obj == NULL)
	return -1;
    tup = PySequence_Tuple(obj);
    if (tup == NULL) {
	Py_DECREF(obj);
	return -1;
    }
    /* Dialog item sequence:
     * classname, caption, id, (x, y, cx, cy), style
     */
    res = PyArg_ParseTuple(tup, "OOi(iiii)i:dialog item",
			   &classname_obj, &caption_obj, &id,
			   &x, &y, &cx, &cy, &style);
    Py_DECREF(tup); Py_DECREF(obj);
    tup = obj = NULL;
    if (!res)
	return -1;

    classname = caption = NULL;
    if (PyString_Check(classname_obj))
	classname = PyString_AsString(classname_obj);
    else if (PyInt_Check(classname_obj))
	predefined = PyInt_AsLong(classname_obj);
    else {
	PyErr_SetString(PyExc_TypeError, "item class; string or int is required");
	return -1;
    }
    if (PyString_Check(caption_obj))
	caption = PyString_AsString(caption_obj);

    return block_len + dlg_item(data,
				style, x, y, cx, cy, id,
				classname, predefined, caption);
}

PyObject *MAKE_DIALOG_TEMPLATE(PyObject *seq)
{
    PyObject *tup, *templ = NULL, *obj = NULL;
    PyObject *menuname_obj, *fontname_obj, *classname_obj;
    char *caption, *menuname, *classname, *fontname;
    int x, y, cx, cy, style, points;
    int num_items;
    int block_len, i;
    char *data;

    if (!PySequence_Check(seq)) {
	PyErr_SetString(PyExc_TypeError, "sequence is required");
	return NULL;
    }
    /* First sequence is dialog header:
     * caption, (x, y, cx, cy), style, menuname, fontname, classname
     */
    obj = PySequence_GetItem(seq, 0);
    if (obj == NULL)
	return NULL;
    tup = PySequence_Tuple(obj);
    if (tup == NULL)
	goto error;
    if (!PyArg_ParseTuple(tup, "s(iiii)iOOO:dialog header",
			  &caption, &x, &y, &cx, &cy, &style,
			  &menuname_obj, &fontname_obj, &classname_obj))
	goto error;
    Py_DECREF(tup); Py_DECREF(obj);
    tup = obj = NULL;

    menuname = classname = fontname = NULL;
    if (PyString_Check(menuname_obj))
	menuname = PyString_AsString(menuname_obj);
    if (PyString_Check(classname_obj))
	classname = PyString_AsString(classname_obj);
    if (PyTuple_Check(fontname_obj)
	&& !PyArg_ParseTuple(fontname_obj, "is:font desc", &points, &fontname))
	    goto error;

    num_items = PySequence_Size(seq) - 1;
    /* First pass - calculate allocation size
     */
    block_len = dlg_header(NULL, style, x, y, cx, cy, num_items,
			   menuname, classname, caption, points, fontname);
    for (i = 0; i < num_items; i++) {
	block_len = parse_dlg_item(NULL, block_len, seq, i);
	if (block_len < 0)
	    return NULL;
    }

    /* Now allocate the buffer for the template
     */
    data = malloc(block_len);
    if (data == NULL)
	return PyErr_NoMemory();

    block_len = dlg_header(data, style, x, y, cx, cy, num_items,
			   menuname, classname, caption, points, fontname);
    for (i = 0; i < num_items; i++) {
	block_len = parse_dlg_item(data, block_len, seq, i);
	if (block_len < 0)
	    return NULL;
    }
    templ = PyCObject_FromVoidPtr(data, free);
	
 error:
    Py_XDECREF(tup);
    Py_XDECREF(obj);
    return templ;
}

int conv_dlgtempl(char *name, PyObject *obj, void **arg)
{
    char msg[128];

    if (PyCObject_Check(obj))
	*arg = PyCObject_AsVoidPtr(obj);
    else {
	sprintf(msg, "%s must be a dialog template", name);
        PyErr_SetString(PyExc_ValueError, msg);
	return 0;
    }
    return 1;
}
